from allennlp.models.encoder_decoders.simple_seq2seq import SimpleSeq2Seq
from allennlp.models.encoder_decoders.composed_seq2seq import ComposedSeq2Seq
from allennlp.models.encoder_decoders.copynet_seq2seq import CopyNetSeq2Seq
