from allennlp.modules.language_model_heads.language_model_head import LanguageModelHead
from allennlp.modules.language_model_heads.bert import BertLanguageModelHead
from allennlp.modules.language_model_heads.gpt2 import Gpt2LanguageModelHead
