class Events:
    TRAINING_START = "TRAINING_START"

    EPOCH_START = "EPOCH_START"

    BATCH_START = "BATCH_START"

    FORWARD = "FORWARD"

    BACKWARD = "BACKWARD"

    BATCH_END = "BATCH_END"

    VALIDATE = "VALIDATE"

    EPOCH_END = "EPOCH_END"

    TRAINING_END = "TRAINING_END"

    ERROR = "ERROR"
